"""
data_mov.py: data movement instructions.
"""
from assembler.errors import check_num_args, StackFull
from assembler.tokens import Instruction, Register, Address


class Fld(Instruction):
    """
        <instr>
             fld
        </instr>
        <syntax>
            fld con
        </syntax>
        <descr>
            loads value onto stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if vm.is_FP_stack_full():
            raise StackFull(line_num)
        # if isinstance(ops[0], FloatTok):
        #     vm.push_to_Float_Stack(ops[0].get_val(line_num))
        vm.push_to_Float_Stack(ops[0].get_val(line_num))


class Fst(Instruction):
    """
        <instr>
             fst
        </instr>
        <syntax>
            fst con
        </syntax>
        <descr>
            stores value from top of stack
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if vm.is_FP_stack_full():
            raise StackFull(line_num)
        ops[0].set_val(vm.registers['ST0'], line_num)


class Mov(Instruction):
    """
        <instr>
             mov
        </instr>
        <syntax>
            MOV reg, reg
            MOV reg, con
            MOV reg, mem
            MOV mem, reg
            MOV mem, mem
        </syntax>
        <descr>
            Copies the value of op2 to the location mentioned in op1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        ops[0].set_val(ops[1].get_val(line_num), line_num)
        if isinstance(ops[0], Register):
            vm.changes.add(ops[0].get_nm())
        elif isinstance(ops[0], Address):
            vm.changes.add(f'MEM{ops[0].get_mem_addr(line_num)}')


class Pop(Instruction):
    """
        <instr>
             pop
        </instr>
        <syntax>
            POP reg
            POP mem
        </syntax>
        <descr>
            POPS the topmost value out of the stack.
            Decrements the stack pointer.
            Can move the stack value to a memory location or register.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.inc_sp(line_num)
        check_num_args("POP", ops, 1, line_num)
        val = int(vm.stack[hex(vm.get_sp()).split('x')[-1].upper()])
        ops[0].set_val(val, line_num)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = vm.empty_cell()


class Push(Instruction):
    """
        <instr>
             push
        </instr>
        <syntax>
            PUSH reg
            PUSH con
            PUSH mem
        </syntax>
        <descr>
            PUSHES the value into the stack with reference to the stack
            pointer position (ESP). Increments the stack pointer automatically,
            everytime a PUSH is called. Callable to store a memory value,
            register value, and constant value to the stack.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.dec_sp(line_num)
        check_num_args("PUSH", ops, 1, line_num)
        vm.stack[hex(vm.get_sp() +
                     1).split('x')[-1].upper()] = ops[0].get_val(line_num)


class Lea(Instruction):
    """
        <instr>
             lea
        </instr>
        <syntax>
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("LEA", ops, 2, line_num)
        # TBD!



'''
Arithmetic and Logical instructions for
RISC-V
'''
import operator as opfunc

from assembler.errors import check_num_args, IncorrectImmLength
from assembler.tokens import Instruction, MAX_INT
# from assembler.ops_check import one_op_arith
from .argument_check import check_reg_only, check_immediate_three
from .argument_check import check_immediate_two


def three_op_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    ops[0].set_val(check_overflow(operator(ops[1].get_val(line_num),
                   ops[2].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def three_op_arith_immediate(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    ops[0].set_val(check_overflow(operator(ops[1].get_val(line_num),
                   ops[2].get_val(line_num)), vm), line_num)
    vm.changes.add(ops[0].get_nm())


def get_three_ops(instr, ops, line_num):
    '''
    Function to grab the values in registers.
    Put in a separate so that we don't have to write out the checks
    all the time.

    '''
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


def get_three_ops_imm(instr, ops, line_num):
    '''
    Same concept as the function above.
    '''
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


def get_two_op_imm(instr, ops, line_num):
    """
    Again, same as above, except for reg, imm format
    """
    check_num_args(instr, ops, 2, line_num)
    check_immediate_two(instr, ops, line_num)
    return (ops[0], ops[1])


def check_overflow(val, vm):
    '''
    To emulate the wraparound that occurs when a number
    has too many bits to represent in machine code.

    '''
    if (val > MAX_INT):
        val = val - MAX_INT + 1
    return val


class Add(Instruction):
    """
        <instr>
             ADD
        </instr>
        <syntax>
            ADD reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.add, line_num)


class Addi(Instruction):
    """
        <instr>
             ADDI
        </instr>
        <syntax>
            ADDI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.add, line_num)


class Sub(Instruction):
    """
        <instr>
            SUB
        </instr>
        <syntax>
            SUB reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


class Mul(Instruction):
    """
        <instr>
            MUL
        </instr>
        <syntax>
            MUL reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.mul, line_num)


class And(Instruction):
    """
        <instr>
            AND
        </instr>
        <syntax>
            AND reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.and_, line_num)


class Andi(Instruction):
    """
        <instr>
            ANDI
        </instr>
        <syntax>
            ANDI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.and_, line_num)


class Or(Instruction):
    """
        <instr>
            OR
        </instr>
        <syntax>
            OR reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)


class Ori(Instruction):
    """
        <instr>
            ORI
        </instr>
        <syntax>
            ORI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.or_, line_num)


class Xor(Instruction):
    """
        <instr>
            XOR
        </instr>
        <syntax>
            XOR reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.xor, line_num)


class Xori(Instruction):
    """
        <instr>
            XORI
        </instr>
        <syntax>
            XORI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.xor, line_num)


class Srl(Instruction):
    """
        <instr>
            SRL
        </instr>
        <syntax>
            SRL reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        fixed_op2 = ops[2].get_val(line_num) % 32
        ops[0].set_val(check_overflow(opfunc.rshift(ops[1].get_val(line_num),
                       fixed_op2), vm), line_num)
        vm.changes.add(ops[0].get_nm())


class Srli(Instruction):
    """
        <instr>
            SRLI
        </instr>
        <syntax>
            SRLI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.rshift, line_num)


class Sll(Instruction):
    """
        <instr>
            SLL
        </instr>
        <syntax>
            SLL reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        fixed_op2 = ops[2].get_val(line_num) % 32
        ops[0].set_val(check_overflow(opfunc.lshift(ops[1].get_val(line_num),
                       fixed_op2), vm), line_num)
        vm.changes.add(ops[0].get_nm())


class Slli(Instruction):
    """
        <instr>
            SLLI
        </instr>
        <syntax>
            SLLI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.lshift, line_num)


class Slt(Instruction):
    """
        <instr>
            SLT
        </instr>
        <syntax>
            SLT reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        if (op2.get_val(line_num) - op3.get_val(line_num)) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sltu(Instruction):
    """
        <instr>
            SLTU
        </instr>
        <syntax>
            SLTU reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        abs_op2 = abs(op2.get_val(line_num))
        abs_op3 = abs(op3.get_val(line_num))
        if (abs_op2 - abs_op3) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Slti(Instruction):
    """
        <instr>
            SLTI
        </instr>
        <syntax>
            SLTI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        if (op2.get_val(line_num) - op3.get_val(line_num)) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sltiu(Instruction):
    """
        <instr>
            SLTU
        </instr>
        <syntax>
            SLTU reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        abs_op2 = abs(op2.get_val(line_num))
        abs_op3 = abs(op3.get_val(line_num))
        if (abs_op2 - abs_op3) < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Sra(Instruction):
    """
        <instr>
            SRA
        </instr>
        <syntax>
            SRA reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        bin_str_op2 = bin(abs(op2.get_val(line_num)))[2:]
        if op2.get_val(line_num) >= 0:
            sign = '0'
        else:
            sign = '1'
        signed_str = sign * op3.get_val(line_num)
        shifted_str = signed_str + bin_str_op2[: - op3.get_val(line_num)]
        op1.set_val(int(shifted_str, 2), line_num)
        vm.changes.add(op1.get_nm())


class Srai(Instruction):
    """
        <instr>
            SRAI
        </instr>
        <syntax>
            SRAI reg, reg, imm
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        bin_str_op2 = bin(abs(op2.get_val(line_num)))[2:]
        if op2.get_val(line_num) >= 0:
            sign = '0'
        else:
            sign = '1'
        signed_str = sign * op3.get_val(line_num)
        shifted_str = signed_str + bin_str_op2[: - op3.get_val(line_num)]
        op1.set_val(int(shifted_str, 2), line_num)
        vm.changes.add(op1.get_nm())


class Div(Instruction):
    """
    <instr>
        DIV
    </instr>
    <syntax>
        DIV reg, reg, reg
    </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.floordiv, line_num)


class Divu(Instruction):
    """
    <instr>
        DIVU
    </instr>
    <syntax>
        DIVU reg, reg, reg
    </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        ops[0].set_val(
            check_overflow(
                opfunc.floordiv(
                    abs(ops[1].get_val(line_num)),
                    abs(ops[2].get_val(line_num))
                ), vm
            ),
            line_num
        )
        vm.changes.add(ops[0].get_nm())


class Rem(Instruction):
    """
    <instr>
        REM
    </instr>
    <syntax>
        REM reg, reg, reg
    </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.mod, line_num)


class Remu(Instruction):
    """
    <instr>
        REMU
    </instr>
    <syntax>
        REMU reg, reg, reg
    </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 3, line_num)
        check_reg_only(self.name, ops, line_num)
        ops[0].set_val(
            check_overflow(
                opfunc.mod(
                       abs(ops[1].get_val(line_num)),
                       abs(ops[2].get_val(line_num))
                ), vm),
            line_num
        )
        vm.changes.add(ops[0].get_nm())


class Lui(Instruction):
    """
    <instr>
        LUI
    </instr>
    <syntax>
        LUI reg, imm
    </syntax>
    <desc>
        Loads a constant (EXPECTED TO BE 20 BITS MAX)
        That's been shifted left by 12 bits
    </desc>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2) = get_two_op_imm(self.get_nm(), ops, line_num)
        if op2.get_val(line_num) > 1048576:
            raise IncorrectImmLength(op2.get_val(line_num), line_num)
        # print(op2.get_val(line_num))
        op1.set_val(check_overflow(opfunc.lshift(op2.get_val(line_num),
                    12), vm), line_num)
        # print(op1.get_val(line_num))
        vm.changes.add(op1.get_nm())


'''
# class Auipc(Instruction):
    """
    <instr>
        AUIPC
    </instr>
    <syntax>
        AUIPC reg, reg, reg
    </syntax>
    """
    # Does lui, and then adds that value to the PC
    # instead of loading to a register.

# class Mulh(Instruction):
    """
    <instr>
        MULH
    </instr>
    <syntax>
        MULH reg, reg, reg
    </syntax>
    """


# class Mulhu(Instruction):
    """
    <instr>
        MULHU
    </instr>
    <syntax>
        MULHU reg, reg, reg
    </syntax>
    """


# class Mulhsu(Instruction):
    """
    <instr>
        MULHSU
    </instr>
    <syntax>
        MULHSU reg, reg, reg
    </syntax>
    """
'''



"""
control_flow.py: control flow instructions,
    plus Exceptions to signal break in flow.

"""

from assembler.errors import check_num_args, InvalidArgument, OutofBounds
from assembler.tokens import Instruction, Register, IntegerTok
from assembler.flowbreak import Jump
from assembler.ops_check import get_one_op
from .argument_check import check_reg_only, check_immediate_three


def get_three_ops(instr, ops, line_num):
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


def get_three_ops_imm(instr, ops, line_num):
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    return (ops[0], ops[1], ops[2])


class Slt(Instruction):
    """
        <instr>
             slt
        </instr>
        <syntax>
            SLT reg, reg, reg
        </syntax>
        <descr>
            Compares op2 and op3, and sets (right now) the SF and ZF flags.
            It is not clear at this moment how to
            treat the OF and CF flags in Python,
            since Python integer arithmetic never carries or overflows!
            Store the result of SF flag into op1
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops(self.get_nm(), ops, line_num)
        res = op2.get_val(line_num) - op3.get_val(line_num)
        if res < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Slti(Instruction):
    """
        <instr>
             slti
        </instr>
        <syntax>
            SLTI reg, con, reg
            SLTI reg, reg, con
        </syntax>
        <descr>
            Compares op2 and op3, and sets (right now) the SF and ZF flags.
            It is not clear at this moment how to
            treat the OF and CF flags in Python,
            since Python integer arithmetic never carries or overflows!
            Store the result of SF flag into op1
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2, op3) = get_three_ops_imm(self.get_nm(), ops, line_num)
        res = op2.get_val(line_num) - op3.get_val(line_num)
        if res < 0:
            op1.set_val(1, line_num)
        else:
            op1.set_val(0, line_num)
        vm.changes.add(op1.get_nm())


class Jmp(Instruction):
    """
        <instr>
            J
        </instr>
        <syntax>
            J lbl
            J loc
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if isinstance(target, IntegerTok):
            raise Jump(str(target.get_val(line_num)))
        else:
            raise Jump(target.name)


class Jal(Instruction):
    """
        <instr>
            JAL
        </instr>
        <syntax>
            JAL loc
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        raise Jump(str(target.get_val(line_num)))


class Jr(Instruction):
    """
        <instr>
            Jr
        </instr>
        <syntax>
            Jr reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        raise Jump(str(target.get_val(line_num)))


class Beq(Instruction):
    """
        <instr>
             BEQ
        </instr>
        <syntax>
            BEQ reg, reg, con
        </syntax>
        <descr>
            Jumps if registers are equal.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("BEQ", ops, 3, line_num)
        disp = 0
        if isinstance(ops[2], IntegerTok):
            disp = ops[2].get_val(line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)
        val_one, val_two = (0, 0)
        if isinstance(ops[0], Register):
            val_one = ops[0].get_val(line_num)
            if isinstance(ops[1], Register):
                val_two = ops[1].get_val(line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            InvalidArgument(ops[0].get_nm(), line_num)
        if val_one == val_two:
            current_ip = vm.get_ip()
            if current_ip + disp * 4 >= 0:
                vm.set_ip(current_ip + disp * 4)
            else:
                raise OutofBounds(line_num)


class Bne(Instruction):
    """
        <instr>
             BNE
        </instr>
        <syntax>
            BNE reg, reg, con
        </syntax>
        <descr>
            Jumps if registers are equal.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("BNE", ops, 3, line_num)
        disp = 0
        if isinstance(ops[2], IntegerTok):
            disp = ops[2].get_val(line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)
        val_one, val_two = (0, 0)
        if isinstance(ops[0], Register):
            val_one = ops[0].get_val(line_num)
            if isinstance(ops[1], Register):
                val_two = ops[1].get_val(line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            InvalidArgument(ops[0].get_nm(), line_num)
        if val_one != val_two:
            current_ip = vm.get_ip()
            if current_ip + disp * 4 >= 0:
                vm.set_ip(current_ip + disp * 4)
            else:
                raise OutofBounds(line_num)



"""
data_mov.py: data movement instructions.
"""
from assembler.errors import check_num_args, InvalidArgument
from assembler.tokens import Instruction, Register, RegAddress


class Load(Instruction):
    """
        <instr>
            LW
        </instr>
        <syntax>
            LW rd, imm(rs1)
        </syntax>
        <descr>
            Load value from memory into rd.
            R[rd] = M_4B[ R[rs1] + sext(imm) ]
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        if isinstance(ops[0], Register):
            if isinstance(ops[1], RegAddress):
                ops[0].set_val(ops[1].get_val(line_num), line_num)
                vm.changes.add(ops[0].get_nm())
            else:
                raise InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Store(Instruction):
    """
        <instr>
            SW
        </instr>
        <syntax>
            SW rs2, imm(rs1)
        </syntax>
        <descr>
            Store the value in rs2 into memory.
            M_4B[ R[rs1] + sext(imm) ] = R[rs2]
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        if isinstance(ops[0], Register):
            if isinstance(ops[1], RegAddress):
                ops[1].set_val(ops[0].get_val(line_num), line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)



"""
interrupts.py: data movement instructions.
"""

from assembler.errors import check_num_args, InvalidOperand
from assembler.errors import ExitProg, UnknownInt
from assembler.tokens import Instruction, IntegerTok

EAX = 'EAX'


def read_key(vm, msg=None):
    # we are faking 'reading' from the keyboard
    c = vm.ret_str[vm.nxt_key]
    vm.nxt_key = (vm.nxt_key + 1) % len(vm.ret_str)
    vm.registers[EAX] = ord(c)
    return ""


def exit_prog(vm, msg):
    raise ExitProg(msg)


int_vectors = {
    22: {0: read_key},
    32: {0: exit_prog},
}


class Interrupt(Instruction):
    """
        <instr>
             int
        </instr>
        <syntax>
            INT con
        </syntax>
        <descr>
            The behavior of INT depends on both its "con" operand
            as well as the value of the EAX register. See the descriptions
            of specific interrupt commands below.
            We will build various "interrupt" handlers as needed.
            At present, we only have two:
                INT 22, with EAX set to 0, to get a key from
            the keyboard. And we only pretend the key is from the keyboard,
            since we are running on the Internet, and can't read the user's
            keyboard.
            And INT 32, to exit the program. There should be a 0 in EAX.
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 1, line_num)
        if type(ops[0]) != IntegerTok:
            raise InvalidOperand(str(ops[0]), line_num)
        try:
            interrupt_class = int_vectors[ops[0].get_val(line_num)]
            interrupt_handler = interrupt_class[int(vm.registers[EAX])]
        except KeyError:
            raise UnknownInt(str(ops[0].get_val(line_num)) + ": "
                             + vm.registers[EAX], line_num)
        c = interrupt_handler(vm, self.get_nm())
        return str(c)



import json
import os
import sys

from IPython.utils.tempdir import TemporaryDirectory
from jupyter_client.kernelspec import KernelSpecManager


kernel_name = "riscv"
kernel_json = {
    "argv": [sys.executable, "-m", "kernels." + kernel_name,
             "-f", "{connection_file}"],
    "display_name": kernel_name.capitalize()
}


def install_kernel(user=True, prefix=None):
    with TemporaryDirectory() as tempdir:
        os.chmod(tempdir, 0o755)  # Starts off as 700, not user readable
        with open(os.path.join(tempdir, "kernel.json"), "w") as f:
            json.dump(kernel_json, f, sort_keys=True)
        print("Installing", kernel_name, "kernel...")
        KernelSpecManager().install_kernel_spec(tempdir, kernel_name, user,
                                                replace=True, prefix=prefix)
        print(kernel_name.capitalize(), "kernel installation complete")


if __name__ == '__main__':
    install_kernel()



"""
data_mov.py: data movement instructions.
"""
from assembler.errors import check_num_args, InvalidArgument
from assembler.tokens import Instruction, Register, RegAddress
from .argument_check import check_even_register

# for floating point to binary and back
import struct

# 64 bits
getBin = lambda x: x > 0 and str(bin(x))[2:] or "-" + str(bin(x))[3:] # noqa


def h_to_b64(value):
    return "0" + hex(int(value, 2))


def f_to_b64(value):
    if (value == 0):
        return "0"*64
    val = struct.unpack('q', struct.pack('d', value))[0]
    return "0" + getBin(val)


def b_to_f64(value):
    if (value == "0"*64):
        return 0.0
    hx = hex(int(value, 2))
    return struct.unpack("d", struct.pack("q", int(hx, 16)))[0]


class Loadc(Instruction):
    """
        <instr>
             LWC
        </instr>
        <syntax>
            LWC reg, reg
            LWC reg, disp(reg)
        </syntax>
        <descr>
            Copies the value of op2 to the location mentioned in op1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        if isinstance(ops[0], Register):
            check_even_register(ops[0], line_num)
            if isinstance(ops[1], RegAddress):
                # if (float(ops[1].get_val(line_num)) > float(2 ** 22)):
                #     raise TooBigForSingle(
                #     str(float(ops[1].get_val(line_num))))
                ops[0].set_val(float(ops[1].get_val(line_num)), line_num)
                vm.changes.add(ops[0].get_nm())
            else:
                raise InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class Storec(Instruction):
    """
        <instr>
             SWC
        </instr>
        <syntax>
            SWC reg, reg
            SWC reg, disp(reg)
        </syntax>
        <descr>
            Copies the value of op2 to the location mentioned in op1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num)
        if isinstance(ops[0], Register):
            check_even_register(ops[0], line_num)
            if isinstance(ops[1], RegAddress):
                # if (float(ops[0].get_val(line_num)) > float(2 ** 22)):
                #     raise TooBigForSingle(
                #     str(float(ops[0].get_val(line_num))))
                ops[1].set_val(float(ops[0].get_val(line_num)), line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class LoadDouble(Instruction):
    """
        <instr>
            LDC
        </instr>
        <syntax>
            LDC reg, reg
            LDC reg, disp(reg)
        </syntax>
        <descr>
            Copies the value of op2 to the location mentioned in op1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num, type_ins=1)
        if isinstance(ops[0], Register):
            check_even_register(ops[0], line_num)
            if isinstance(ops[1], RegAddress):
                bin_string = f_to_b64(ops[1].get_val(line_num))
                # split the string into two
                first_half = bin_string[:32]
                second_half = bin_string[32:]
                # since we're storing in 2 registers
                # we need to know which register comes next...
                reg_number = int(ops[0].get_nm()[1:])
                next_reg = "F" + str(reg_number + 1)

                ops[0].set_val(first_half, line_num)
                vm.registers[next_reg] = second_half
                vm.changes.add(ops[0].get_nm())
                vm.changes.add(next_reg)
            else:
                raise InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)


class StoreDouble(Instruction):
    """
        <instr>
             SDC
        </instr>
        <syntax>
            SDC reg, reg
            SDC reg, disp(reg)
        </syntax>
        <descr>
            Copies the value of op2 to the location mentioned in op1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.get_nm(), ops, 2, line_num, type_ins=1)
        if isinstance(ops[0], Register):
            check_even_register(ops[0], line_num)
            if isinstance(ops[1], RegAddress):
                reg_number = int(ops[0].get_nm()[1:])
                curr_reg = "F" + str(reg_number + 0)
                next_reg = "F" + str(reg_number + 1)
                first_half = vm.registers[curr_reg]
                second_half = vm.registers[next_reg]
                full_b = first_half + second_half
                v = b_to_f64(full_b)
                ops[1].set_val(v, line_num)

                # deprecated code below
                # print(ops[0])
                # first_half = ops[0].get_val(line_num)
                # reg_name = "F" + str(int(ops[0].get_nm()[1:]))
                # second_half = vm.registers[reg_name]
                # full_b = first_half + second_half
                # print("full_b", full_b)
                # v = b_to_f64(full_b)
                # ops[1].set_val(v)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)



from .arithmetic import Add, Sub, Mul, Div_S, Div_U
from .arithmetic import Rem_S, Rem_U, And, Or, Xor
from .arithmetic import Shl, Shr_S, Shr_U, Rotl, Rotr
from .arithmetic import Clz, Ctz, Popcnt, Eqz
from .data_mov import Global_set, Local_set, Store_global, Store_local
from .data_mov import Global_get, Local_get, Store_const
from assembler.tokens import DataType

key_words = {
    # data types
    'i32': DataType('i32'),

    # arithmetic
    'i32.add': Add('.add'),
    'i32.sub': Sub('.sub'),
    'i32.mul': Mul('.mul'),
    'i32.div_s': Div_S('.div_s'),
    'i32.div_u': Div_U('.div_u'),
    'i32.rem_s': Rem_S('.rem_s'),
    'i32.rem_u': Rem_U('.rem_u'),
    'i32.and': And('.and'),
    'i32.or': Or('.or'),
    'i32.xor': Xor('.xor'),
    'i32.shl': Shl('.shl'),
    'i32.shr_s': Shr_S('.shr_s'),
    'i32.shr_u': Shr_U('.shr_u'),
    'i32.rotl': Rotl('.rotl'),
    'i32.rotr': Rotr('.rotr'),
    'i32.clz': Clz('.clz'),
    'i32.ctz': Ctz('.ctz'),
    'i32.popcnt': Popcnt('.popcnt'),
    'i32.eqz': Eqz('.eqz'),

    # data movement
    'global.get': Global_get('global.get'),
    'local.get': Local_get('local.get'),
    'global.set': Global_set('global.set'),
    'local.set': Local_set('local.set'),
    'global': Store_global('global'),
    'local': Store_local('local'),
    'i32.const': Store_const('i32.const')
}



"""
tokens.py: contains classes we tokenize into.
"""

# for floating point to binary and back
import struct
import binascii
import sys

from abc import abstractmethod

from .errors import InvalidMemLoc, RegUnwritable, IntOutOfRng
from .errors import UnknownName, InvalidArgument
from .errors import NotSettable, UnknownLabel, LabelNotSettable
from .errors import TooBigForSingle, TooBigForDouble

BITS = 32   # we are on a 32-bit machine
MAX_INT = (2**(BITS-1)) - 1
MIN_INT = -(2**(BITS-1))

MAX_FLOAT = sys.float_info.max
MIN_FLOAT = sys.float_info.min

VALS = 0
MEM_LOC = 1


def add_debug(s, vm):
    vm.debug += (s + "\n")


# 32 bits
def float_to_hex(f):
    return hex(struct.unpack('<I', struct.pack('<f', f))[0])


def hex_to_float(h):
    h2 = h[2:]
    h2 = binascii.unhexlify(h2)
    return struct.unpack('>f', h2)[0]


# 64 bits
getBin = lambda x: x > 0 and str(bin(x))[2:] or "-" + str(bin(x))[3:] # noqa


def h_to_b64(value):
    return "0" + hex(int(value, 2))


def f_to_b64(value):
    if (value == 0):
        return "0"*64
    val = struct.unpack('q', struct.pack('d', value))[0]
    return "0" + getBin(val)


def b_to_f64(value):
    if (value == "0"*64):
        return 0.0
    hx = hex(int(value, 2))
    return struct.unpack("d", struct.pack("q", int(hx, 16)))[0]


class Token:
    def __init__(self, name, val=0):
        self.name = name
        self.value = val

    def __str__(self):
        return str(self.name)

    def set_val(self, val):
        raise NotSettable(str(self))

    def get_val(self, line_num=None):
        return self.value

    def get_nm(self):
        return self.name


class Section(Token):
    def __init__(self, name):
        super().__init__(name)


class OpenBracket(Token):
    def __init__(self):
        super().__init__("[")


class CloseBracket(Token):
    def __init__(self):
        super().__init__("]")


class OpenParen(Token):
    def __init__(self):
        super().__init__("(")


class CloseParen(Token):
    def __init__(self):
        super().__init__(")")


class DataType(Token):
    def __init__(self, name):
        super().__init__(name)


class Comma(Token):
    def __init__(self):
        super().__init__(",")


class ConstantSign(Token):
    """
    Class used to differentiate between
    a constant and an offset
    """
    def __init__(self):
        super().__init__("$")


class Instruction(Token):
    """
    Class representing all instructions.
    """
    def __init__(self, name):
        super().__init__(name)

    def __str__(self):
        return str(self.name)

    def f(self, ops, vmachine, line_number):
        return self.fhook(ops, vmachine, line_number)

    @abstractmethod
    def fhook(self, ops, vmachine, line_number):
        pass


class Operand(Token):
    """
    Superclass of all operands.
    """
    def __init__(self, name, val=0):
        super().__init__(name, val)


class IntegerTok(Operand):
    def __init__(self, val=0, con=True):
        if(val > MAX_INT or val < MIN_INT):
            raise IntOutOfRng(str(val))

        super().__init__("Integer", val)
        self.con = con

    def __str__(self):
        return str(self.value)

    def get_val(self, line_num):
        return self.value

    def negate_val(self):
        self.value *= -1


class FloatTok(Operand):
    def __init__(self, data_type=".float", val=0.0):
        self.data_type = data_type
        # do a bit of error checking for precision for the hex value
        if data_type == ".float":
            # if type(val) is float:
            #     temp_hex = float_to_hex(val)
            #     temp_val = hex_to_float(temp_hex)
            #     if (temp_val != val):
            #         raise TooBigForSingle(str(val))
            if type(val) is str:
                temp_float = hex_to_float(val)
                temp_hex = float_to_hex(temp_float)
                if (temp_hex != val):
                    raise TooBigForSingle(str(val))
        elif data_type == ".double":
            if type(val) is float:
                temp_bin = f_to_b64(val)
                temp_float = b_to_f64(temp_bin)
                if (temp_float != val):
                    raise TooBigForDouble(str(val))
            elif type(val) is str:
                temp_float = b_to_f64(val)
                temp_bin2 = f_to_b64(temp_float)
                temp_float2 = b_to_f64(temp_bin2)
                if temp_float != temp_float2:
                    raise TooBigForDouble(str(val))

        super().__init__("Float", val)

    def __str__(self):
        return str(self.get_val())

    def get_type(self):
        return self.data_type

    # self.value is either going to be a float (12.2)
    # or a hexadecimal string ('0x41433333')
    # we need to be able to reconcile the actual value of it
    # if it's a hex string (IEEE 754)
    def get_val(self, line_num):
        if type(self.value) is float:
            return self.value

        # it is single precision as hexadecimal
        # convert the hexadecimal to be a float
        if self.data_type == ".float":
            return float('%.3f' % (hex_to_float(self.value)))
        else:
            return float('%.4f' % (b_to_f64(self.value)))

    def negate_val(self):
        self.value *= -1.0


class StringTok(Token):
    def __init__(self, name):
        super().__init__(name)


class DupTok(Token):
    def __init__(self):
        super().__init__("DUP")


class QuestionTok(Token):
    def __init__(self):
        super().__init__("?")


class PlusTok(Token):
    def __init__(self):
        super().__init__("+")


class MinusTok(Token):
    def __init__(self):
        super().__init__("-")


class Location(Operand):
    """
    Class to give common type to memory and registers.
    Adds set_val(), not possible for ints!
    """
    def __init__(self, name, vm, val=0):
        super().__init__(name, val)
        self.vm = vm

    @abstractmethod
    def set_val(self, val):
        pass


class Address(Location):
    def __init__(self, name, vm, val=0):
        super().__init__(name, vm, val)
        self.mem = vm.memory

    def __str__(self):
        return "[" + str(self.name) + "]"

    def get_val(self, line_num):
        if self.name in self.mem:
            if "." in str(self.mem[self.name]):
                return float(self.mem[self.name])
            return int(self.mem[self.name])
        else:
            return 0

    def set_val(self, val, line_num):
        self.mem[self.name] = val

    def get_mem_addr(self, line_num):
        return self.name


class RegAddress(Address):
    def __init__(self, name, vm, displacement=0, multiplier=1, val=0):
        super().__init__(name, vm, val)
        self.regs = vm.registers
        self.displacement = displacement
        self.multiplier = multiplier

    def get_mem_addr(self, line_num):
        # right now, memory addresses are strings. eeh!
        address = hex(int(self.regs[self.name]) *
                      self.multiplier).split('x')[-1].upper()
        disp = 0
        if isinstance(self.displacement, list):
            for disp_item in self.displacement:
                if isinstance(disp_item, Register):
                    disp += disp_item.get_val(line_num) * \
                            disp_item.get_multiplier()
                else:
                    disp += disp_item
        elif isinstance(self.displacement, Register):
            disp = self.displacement.get_val(line_num)
        elif self.displacement != 0:
            disp = self.displacement
        addr_val = int(self.regs[self.name]) * self.multiplier + disp
        if addr_val < 0:
            raise InvalidMemLoc(str(addr_val), line_num)
        address = hex(addr_val).split('x')[-1].upper()
        return address

    def get_val(self, line_num):
        mem_addr = self.get_mem_addr(line_num)
        if mem_addr in self.mem:
            return self.mem[str(mem_addr)]
        else:
            return 0

    def set_val(self, val, line_num):
        mem_addr = self.get_mem_addr(line_num)
        self.mem[str(mem_addr)] = val


class Register(Location):
    def __init__(self, name, vm, val=0):
        super().__init__(name, vm, val)
        # isFloat = False
        # if name[:2].upper() == 'ST':
        #     isFloat = True
        # if isFloat:
        #     self.registers = vm.fp_stack_registers
        # # if FLOAT ==True:
        # #     self.registers = vm.fp_stack_registers
        # #     #self.val = self.registers[self.name]
        # #     print("registers changed")
        # else:
        #     self.registers = vm.registers
        self.registers = vm.registers
        self.val = self.registers[self.name]
        self.writable = True
        self.multiplier = 1
        if self.name in vm.unwritable:
            self.writable = False

    def __str__(self):
        return str(self.name)

    def get_val(self, line_num):
        if self.name[:2].upper() == "ST" or self.name[0].upper() == "F":
            return float(self.registers[self.name])
        return int(self.registers[self.name])

    def set_val(self, val, line_num):
        if self.writable:
            self.registers[self.name] = val
        else:
            raise RegUnwritable(self.name, line_num)

    def get_multiplier(self):
        return self.multiplier

    def set_multiplier(self, val):
        if val < 0:
            raise InvalidArgument(str(val))
        self.multiplier = val

    def negate_val(self):
        self.val *= -1


class Label(Location):
    """
    Class to hold labels for jumps.
    """
    def __init__(self, name, vm, val=0):
        super().__init__(name, vm, val)
        self.labels = vm.labels
        if self.name not in self.labels and val != 0:
            self.labels[self.name] = val

    def get_val(self, line_num):
        if self.name not in self.labels:
            raise UnknownLabel(self.name)
        else:
            return self.labels[self.name]

    def set_val(self, val):
        raise LabelNotSettable(self.name)


class NewSymbol(Token):
    def __init__(self, name, index=None):
        super().__init__(name)
        self.val = 0

    def get_val(self, line_num):
        return self.val

    def set_val(self, value):
        self.val = value


class Symbol(Location):
    """
    Class to hold symbols such as variable names.
    """
    def __init__(self, name, vm):
        super().__init__(name, vm)
        self.vm = vm
        self.check_nm()

    def check_nm(self):
        if self.name not in self.vm.symbols:
            raise UnknownName(self.name)

    def set_val(self, val):
        self.check_nm()
        self.vm.symbols[self.name] = val

    def get_val(self, line_num):
        self.check_nm()
        add_debug("Symbol " + self.name + " = "
                  + str(self.vm.symbols[self.name]),
                  self.vm)
        return self.vm.symbols[self.name]



"""
control_flow.py: control flow instructions,

"""

from assembler.errors import check_num_args, OutofBounds, InvalidArgument
from assembler.tokens import Instruction, Register, IntegerTok
from assembler.flowbreak import Jump
from assembler.ops_check import get_one_op
from .argument_check import check_immediate_two


def get_two_op_imm(instr, ops, line_num):
    check_num_args(instr, ops, 2, line_num)
    check_immediate_two(instr, ops, line_num)
    return (ops[0], ops[1])


class Jr(Instruction):
    """
        <instr>
            JR
        </instr>
        <syntax>
            JR reg
        </syntax>
        <descr>
            Jump to address.
            PC = R[rs1]
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        raise Jump(str(target.get_val(line_num)))


class Jal(Instruction):
    """
        <instr>
            JAL
        </instr>
        <syntax>
            JAL rd, imm
        </syntax>
        <descr>
            Jump to address and place return address in GPR.
            R[rd] = PC + 4; PC = PC + sext(imm)
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2) = get_two_op_imm(self.get_nm(), ops, line_num)
        current_ip = vm.get_ip()
        target = current_ip + op2.get_val(line_num)
        print(op2.get_val(line_num))
        print("curr", current_ip, 'tar', target)

        op1.set_val(current_ip + 4, line_num)
        print(current_ip + 4)
        vm.changes.add(op1.get_nm())
        raise Jump(str(target))

# The original implementation of JAL was pretty off.
# I've corrected it, but it needs to be tested some more.


class Jalr(Instruction):
    """
        <instr>
            JALR
        </instr>
        <syntax>
            JALR rd, rs1, imm
        </syntax>
        <descr>
            Jump to address and place return address in GPR
            R[rd] = PC + 4; PC = ( R[rs1] + sext(imm) ) & 0xfffffffe
            The and makes it so that the least significant bit is 0
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        current_ip = vm.get_ip()

        target = ops[1] + ops[2]
        ops[0].set_val(current_ip + 4, line_num)
        vm.changes.add(ops[0].get_nm())
        raise Jump(str(target))
# I need to find a better way to zero out the LSB. I think I
# will be converting into binary form of string, slicing and
# then converting back. It's the converting back that I have
# conflicts on.


class Beq(Instruction):
    """
        <instr>
            BEQ
        </instr>
        <syntax>
            BEQ rs1, rs2, imm
        </syntax>
        <descr>
            Branch if 2 GPRs are equal.
            PC = ( R[rs1] == R[rs2] ) ? PC + sext(imm) : PC + 4
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("BEQ", ops, 3, line_num)
        disp = 0
        if isinstance(ops[2], IntegerTok):
            disp = ops[2].get_val(line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)
        val_one, val_two = (0, 0)
        if isinstance(ops[0], Register):
            val_one = ops[0].get_val(line_num)
            if isinstance(ops[1], Register):
                val_two = ops[1].get_val(line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            InvalidArgument(ops[0].get_nm(), line_num)
        if val_one == val_two:
            current_ip = vm.get_ip()
            if current_ip + disp * 4 >= 0:
                vm.set_ip(current_ip + disp * 4)
            else:
                raise OutofBounds(line_num)


class Bne(Instruction):
    """
        <instr>
            BNE
        </instr>
        <syntax>
            BNE rs1, rs2, imm
        </syntax>
        <descr>
            Branch if 2 GPRs are not equal.
            PC = ( R[rs1] != R[rs2] ) ? PC + sext(imm) : PC + 4
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("BNE", ops, 3, line_num)
        disp = 0
        if isinstance(ops[2], IntegerTok):
            disp = ops[2].get_val(line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)
        val_one, val_two = (0, 0)
        if isinstance(ops[0], Register):
            val_one = ops[0].get_val(line_num)
            if isinstance(ops[1], Register):
                val_two = ops[1].get_val(line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            InvalidArgument(ops[0].get_nm(), line_num)
        if val_one != val_two:
            current_ip = vm.get_ip()
            if current_ip + disp * 4 >= 0:
                vm.set_ip(current_ip + disp * 4)
            else:
                raise OutofBounds(line_num)


class Blt(Instruction):
    """
        <instr>
            BLT
        </instr>
        <syntax>
            BLT rs1, rs2, imm
        </syntax>
        <descr>
            Branch based on signed comparison of two GPRs
            PC = ( R[rs1] <s R[rs2] ) ? PC + sext(imm) : PC + 4
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("BNE", ops, 3, line_num)
        disp = 0
        if isinstance(ops[2], IntegerTok):
            disp = ops[2].get_val(line_num)
        else:
            raise InvalidArgument(ops[0].get_nm(), line_num)
        val_one, val_two = (0, 0)

        if isinstance(ops[0], Register):
            val_one = ops[0].get_val(line_num)
            if isinstance(ops[1], Register):
                val_two = ops[1].get_val(line_num)
            else:
                InvalidArgument(ops[1].get_nm(), line_num)
        else:
            InvalidArgument(ops[0].get_nm(), line_num)

        if val_one < val_two:
            current_ip = vm.get_ip()
            if current_ip + disp * 4 >= 0:
                vm.set_ip(current_ip + disp * 4)
            else:
                raise OutofBounds(line_num)



from .arithmetic import Add, Sub, Addi, Mult, Div
from .arithmetic import Andf, Andi, Orf, Ori, Xor, Nor, Sll, Srl
from .arithmetic import Mflo, Mfhi
from .data_mov import Load, Store
from .control_flow import Slt, Slti, Beq, Bne, Jmp, Jal, Jr
from .interrupts import Syscall
from assembler.tokens import DataType
from .fp_arithmetic import Adds, Subs, Mults, Divs
from .fp_arithmetic import Addd, Subd, Multd, Divd
from .fp_data_mov import Loadc, Storec, LoadDouble, StoreDouble

key_words = {
    # data types
    '.BYTE': DataType('DB'),
    '.WORD': DataType('DW'),
    '.FLOAT': DataType('FL'),
    '.DOUBLE': DataType('DBL'),

    # data movement:
    'LW': Load('LW'),
    'SW': Store('SW'),

    # arithmetic and logic
    'ADD': Add('ADD'),
    'ADDI': Addi('ADDI'),
    'SUB': Sub('SUB'),
    'MULT': Mult('MULT'),
    'DIV': Div('DIV'),
    'MFLO': Mflo('MFLO'),
    'MFHI': Mfhi('MFHI'),
    'AND': Andf('AND'),
    'ANDI': Andi('ANDI'),
    'OR': Orf('OR'),
    'ORI': Ori('ORI'),
    'XOR': Xor('XOR'),
    'NOR': Nor('NOR'),
    'SLL': Sll('SLL'),
    'SRL': Srl('SRL'),


    # control
    'SLT': Slt('SLT'),
    'SLTI': Slti('SLTI'),
    'BEQ': Beq('BEQ'),
    'BNE': Bne('BNE'),
    'J': Jmp('J'),
    'JAL': Jal('JAL'),
    'JR': Jr('JR'),

    # interrupts
    'SYSCALL': Syscall('SYSCALL'),

    # floating point
    'ADD.S': Adds('ADD.S'),
    'SUB.S': Subs('SUB.S'),
    'MULT.S': Mults('MULT.S'),
    'DIV.S': Divs('DIV.S'),

    'ADD.D': Addd("ADD.D"),
    'SUB.D': Subd('SUB.D'),
    'MULT.D': Multd('MULT.D'),
    'DIV.D': Divd('DIV.D'),

    # floating point data mov
    'LWC': Loadc('LWC'),
    'SWC': Storec('SWC'),
    'LDC': LoadDouble('LDC'),
    'SDC': StoreDouble('SDC')
}

op_func_codes = {
    # R-format
    'ADD': ('000000', '100000'),
    'SUB': ('000000', '100010'),
    'MULT': ('000000', '011000'),
    'DIV': ('000000', '011010'),
    'MFLO': ('000000', '010010'),
    'MFHI': ('000000', '010000'),
    'AND': ('000000', '100100'),
    'OR': ('000000', '100101'),
    'XOR': ('000000', '100110'),
    'NOR': ('000000', '100111'),
    'SLL': ('000000', '000000'),
    'SRL': ('000000', '000010'),
    'SLT': ('000000', '101010'),
    'JR': ('000000', '001000'),
    # I-format
    'LW': '100011',
    'SW': '101011',
    'ADDI': '001000',
    'ANDI': '001100',
    'ORI': '001101',
    'SLTI': '001010',
    'BEQ': '000100',
    'BNE': '000101',
    'SYSCALL': '001100',

    # J-format
    'J': '000010',
    'JAL': '000011',

    # FR
    'ADD.S': ('010001'),
    'SUB.S': ('010001', "000001"),
    'MULT.S': ('010001', '000010'),
    'DIV.S': ('010001', '000010'),
    'LWC': ('110001'),
    'SWC': ('111001'),
    'LDC': ('110101'),
    'SDC': ('111101'),

    'ADD.D': ('010001'),
    'SUB.D': ('010001'),
    'MULT.D': ('010001'),
    'DIV.D': ('010001'),
}



"""
fp_arithmetic.py: arithmetic floating point instructions.
"""

from assembler.errors import check_num_args, DivisionZero, InvalidOperand
from assembler.tokens import Instruction, MAX_FLOAT
# from .arithmetic import checkflag
# from assembler.virtual_machine import intel_machine
from .fp_conversions import add, sub, mul, div, fabs, chs
import math


def convert_hex_to_decimal(fhex):
    """
    :param fhex: floating point hexadecimal number in str format
    :return: decimal equivalent of fhex in float format
    Eg : 'a2.4c' -> 162.296875
    """
    mapping = {}
    for i in range(10):
        mapping[str(i)] = i
    for i in range(10, 16):
        mapping[chr(i-10+97)] = i

    flag = 1
    if fhex[0] == '-':
        flag = -1
        fhex = fhex[1:]
    if '.' not in fhex:
        return int(fhex, 16)
    before_point_hex, after_point_hex = fhex.split('.')
    before_point_dec, after_point_dec = 0, 0
    for i in range(len(before_point_hex)):
        map_before_pt_hex = mapping[before_point_hex[i]]
        before_point_dec += map_before_pt_hex*(16**(len(before_point_hex)-i-1))
    for i in range(len(after_point_hex)):
        after_point_dec += mapping[after_point_hex[i]]*(16**(-1*(i+1)))
    res = '{}.{}'.format(str(before_point_dec), str(after_point_dec)[2:])
    return flag*float(res)


def convert_dec_to_hex(fdec):
    """
    :param fdec: floating point decimal number in float format
    :return: hexadecimal equivalent of fdec in str format
    Eg : 162.296875 -> 'a2.4c'
    """
    flag = 1
    if fdec < 0:
        flag = -1
    fdec = flag * fdec
    fdec = str(fdec)
    if '.' not in fdec:
        if flag == -1:
            return '-' + hex(int(fdec))[2:]
        else:
            return hex(int(fdec))[2:]
    before_point_dec, after_point_dec = fdec.split('.')
    before_point_hex = hex(int(before_point_dec))[2:]
    binary = convert_after_point_dec_to_binary('0.' + after_point_dec)
    after_point_hex = ''
    for i in range(0, len(binary), 4):
        after_point_hex += convert_grouped_binary_to_hex(binary[i:i + 4])
    if flag == -1:
        final_hex = '-' + before_point_hex + '.' + after_point_hex
    elif flag == 1:
        final_hex = before_point_hex + '.' + after_point_hex
    return final_hex


def convert_after_point_dec_to_binary(dec):
    """
    :param dec: decimal fraction number in string format
    :return: binary equivalent of dec in str format
    Eg : 0.9 -> '11100110011001100110011001100110'
    """
    a = float(dec)
    binary = ''
    while a > 0.0:
        a = a*2
        binary += str(int(a))
        if a >= 1:
            a -= 1
    return binary


def convert_grouped_binary_to_hex(binary):
    """

    :param binary: group of 4 binary number in string format
    :return hexadecimal equivalent of binary number in string format:
    """
    mapping = {}
    for i in range(10):
        mapping[i] = str(i)
    for i in range(10, 16):
        mapping[i] = chr(i-10+97)
    binary = binary+'0'*(4-len(binary))
    integer = int(binary, 2)
    hexequi = mapping[integer]
    return hexequi


"""
dec_convert method is not being used
"""


def dec_convert(val):
    while val > 1:
        val = val / 10
    return val


def checkflag(val, vm):
    """

    :param val: float
    :param vm:
    :return: val in float

    if overflow happens in that case it sets carry flag to one

    """
    if(val > MAX_FLOAT):
        vm.flags['CF'] = 1
        val = val - MAX_FLOAT+1
    else:
        vm.flags['CF'] = 0
    return val


def two_op_arith(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 2, line_num)
    reg_one, reg_two = [int(x.get_nm()[-1]) for x in ops]
    # first_reg = vm.get_float_stack_register_at_offset(offset1)
    # second_reg = vm.get_float_stack_register_at_offset(offset2)
    # r1 = vm.fp_stack_registers[first_reg]
    # r2 = vm.fp_stack_registers[second_reg]
    # vm.fp_stack_registers[first_reg] = checkflag(operator(r1, r2), vm)
    if reg_one != 0 and reg_two != 0:
        raise InvalidOperand('Neither registers are ST0', line_num)
    r1 = vm.registers[f'ST{reg_one}']
    r2 = vm.registers[f'ST{reg_two}']
    vm.registers[f'ST{reg_one}'] = checkflag(operator(r1, r2), vm)
    # r1.set_val(
    #     checkflag(operator(r1.get_val(line_num),
    #                        r2.get_val(line_num)), vm))


def pop_arith(ops, vm, instr, operator, line_num):
    check_num_args(instr, ops, 2, line_num)
    reg_one, reg_two = [int(x.get_nm()[-1]) for x in ops]
    # first_reg = vm.get_float_stack_register_at_offset(offset1)
    # second_reg = vm.get_float_stack_register_at_offset(offset2)
    # r1 = vm.fp_stack_registers[first_reg]
    # r2 = vm.fp_stack_registers[second_reg]
    # vm.fp_stack_registers[first_reg] = checkflag(operator(r1, r2), vm)
    if reg_two != 0:
        raise InvalidOperand(f'ST{reg_two}', line_num)
    r1 = vm.registers[f'ST{reg_one}']
    r2 = vm.registers[f'ST{reg_two}']
    vm.registers[f'ST{reg_one}'] = checkflag(float(operator(r1, r2)), vm)
    # r1.set_val(
    #     checkflag(operator(r1.get_val(line_num),
    #                        r2.get_val(line_num)), vm))


def one_op_arith(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 1, line_num)
    if operator.__name__ == 'div' and ops[0].get_val(line_num) == 0.0:
        raise DivisionZero(line_num)
    else:
        vm.registers["ST0"] = operator(vm.registers["ST0"],
                                       ops[0].get_val(line_num))


class FAdd(Instruction):
    """
    1 op - adds val to stack top ST(0) and stores value at ST(0)
    2 ops - sets sum  of floating stack ST(i) and floating stack
    ST(j) to floating stack ST(i)
        <instr>
             FADD
        </instr>
        <syntax>
            FADD val
            FADD ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, add, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, add, line_num)


class FaddP(Instruction):
    """
        1 op - adds val to stack top ST(0)
        and stores value at ST(0) and then pops stack
        2 ops - sets sum  of floating stack ST(i)
        and floating stack  ST(j) to floating stack ST(i)
        and then pops stack
            <instr>
                 FADDP
            </instr>
            <syntax>
                FADDP val
                FADDP ST(i), ST(j)
            </syntax>
        """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, add, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, add, line_num)
        vm.pop_from_Float_Stack()


class FSub(Instruction):
    """
    1 op - subtracts val from stack top ST(0) and stores value at ST(0)
    2 ops - sets difference  of floating stack ST(i) and floating stack
    ST(j) to floating stack ST(i)
        <instr>
             FSUB
        </instr>
        <syntax>
            FSUB val
            FSUB ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, sub, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, sub, line_num)


class FSubP(Instruction):
    """
    1 op - subtracts val from stack top ST(0)
    and stores value at ST(0) and then pops the stack
    2 ops - sets difference  of floating stack ST(i) and floating stack
    ST(j) to floating stack ST(i) and then pops the stack
        <instr>
             FSUBP
        </instr>
        <syntax>
            FSUBP val
            FSUBP ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, sub, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, sub, line_num)
        vm.pop_from_Float_Stack()


class FMul(Instruction):
    """
    1 op - multiplies val with stack top ST(0) and stores value at ST(0)
    2 ops - sets product of floating stack ST(i) and floating stack
     ST(j) to floating stack ST(i)
        <instr>
             FMUL
        </instr>
        <syntax>
            FMUL val
            FMUL ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, mul, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, mul, line_num)


class FMulP(Instruction):
    """
    1 op - multiplies val with stack top ST(0)
    and stores value at ST(0) and then pops the stack
    2 ops - sets product of floating stack ST(i) and floating stack
     ST(j) to floating stack ST(i) and then pops the stack
        <instr>
             FMULP
        </instr>
        <syntax>
            FMULP val
            FMULP ST(i), ST(j)
        </syntax>
    """

    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, mul, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, mul, line_num)
        vm.pop_from_Float_Stack()


class FAbs(Instruction):
    """
    sets bit  of floating-point register (FPR) FRB to 0
    and place the results into FPR FRT
        <instr>
                FABS
        </instr>
        <syntax>
            fabs FRT, FRB
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, fabs, line_num)


class FChs(Instruction):
    """
    complements the sign of floating-point register (FPR) FRB
        <instr>
                FCHS
        </instr>
        <syntax>
            fchs FRT
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, chs, line_num)


class FDiv(Instruction):
    """
    1 op - divides stack top ST(0) with val and stores the result at ST(0)
    2 ops - sets the result of dividing floating stack ST(i) by floating stack
     ST(j) to floating stack ST(i)
        <instr>
             FDIV
        </instr>
        <syntax>
            FDIV val
            FDIV ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, div, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, div, line_num)


class FDivP(Instruction):
    """
    1 op - divides stack top ST(0) with val
    and stores the result at ST(0) and pops the stack
    2 ops - sets the result of dividing floating stack ST(i) by floating stack
     ST(j) to floating stack ST(i) and pops the stack
        <instr>
             FDIVP
        </instr>
        <syntax>
            FDIVP val
            FDIVP ST(i), ST(j)
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 1:
            one_op_arith(ops, vm, self.name, div, line_num)
        elif len(ops) == 2:
            two_op_arith(ops, vm, self.name, div, line_num)
        vm.pop_from_Float_Stack()


class FSqrt(Instruction):
    """
    0 op - computes the square root of the source value in the ST(0)
    register and stores the result in ST(0)
        <instr>
             FSQRT
        </instr>
        <syntax>
            FSQRT
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        if len(ops) == 0:
            top_value = vm.pop_from_Float_Stack()
            sqrt = math.sqrt(top_value)
            vm.push_to_Float_Stack(sqrt)



"""
arithmetic.py: arithmetic instructions.
"""

import operator as opfunc
from assembler.errors import DivisionZero
from assembler.tokens import Instruction, MAX_INT


def get_stack_operand(vm):
    position = hex(vm.get_sp()).split('x')[-1].upper()
    val_one = vm.stack[position]
    vm.stack[position] = 0
    return val_one


def get_stack_operands(vm, line_num):
    vm.dec_sp(line_num)
    val_one = get_stack_operand(vm)
    vm.dec_sp(line_num)
    val_two = get_stack_operand(vm)
    return val_one, val_two


def two_op_arith(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    val_one, val_two = get_stack_operands(vm, line_num)
    # vm.dec_sp(line_num)
    position = hex(vm.get_sp()).split('x')[-1].upper()
    vm.stack[position] = operator(val_one, val_two)
    vm.inc_sp(line_num)


def check_overflow(val, vm):
    '''
    To emulate the wraparound that occurs when a number
    has too many bits to represent in machine code.

    '''
    if (val > MAX_INT):
        val = val - MAX_INT + 1
    return val


class Add(Instruction):
    """
        <instr>
             .add
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.add
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.add, line_num)


class Sub(Instruction):
    """
        <instr>
             .sub
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.sub
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.sub, line_num)


class Mul(Instruction):
    """
        <instr>
             .mul
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.mul
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.mul, line_num)


class Div_S(Instruction):
    """
        <instr>
             .div_s
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.div_s
        </syntax>
        <descr>
            This is the division instruction for signed values.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.floordiv, line_num)


class Div_U(Instruction):
    """
        <instr>
             .div_u
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.div_u
        </syntax>
        <descr>
            This is the division instruction for unsigned values.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        if val_two == 0:
            raise DivisionZero(line_num)
        result = opfunc.floordiv(abs(val_one), abs(val_two))
        check_overflow(result, vm)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Rem_S(Instruction):
    """
        <instr>
             .rem_s
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.rem_s
        </syntax>
        <descr>
            This is the remainder instruction for signed values.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.mod, line_num)


class Rem_U(Instruction):
    """
        <instr>
             .rem_u
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.rem_u
        </syntax>
        <descr>
            This is the remainder instruction for unsigned values.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        result = opfunc.mod(abs(val_one), abs(val_two))
        check_overflow(result, vm)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class And(Instruction):
    """
        <instr>
             .and
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.and
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.and_, line_num)


class Or(Instruction):
    """
        <instr>
             .or
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.or
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.or_, line_num)


class Xor(Instruction):
    """
        <instr>
             .xor
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.xor
        </syntax>
        <descr>
            $lhs and $rhs are example variables
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        two_op_arith(ops, vm, self.name, opfunc.xor, line_num)


class Shl(Instruction):
    """
        <instr>
             .shl
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.shl
        </syntax>
        <descr>
            Integer Shift Left
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        val_two = val_two % 32  # only for i32
        result = opfunc.lshift(val_one, val_two)
        check_overflow(result, vm)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Shr_S(Instruction):
    """
        <instr>
             .shr_s
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.shr_s
        </syntax>
        <descr>
            Integer Shift Right Signed
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        val_two = val_two % 32  # only for i32
        result = opfunc.rshift(val_one, val_two)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Shr_U(Instruction):
    """
        <instr>
             .shr_u
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.shr_u
        </syntax>
        <descr>
            Integer Shift Right Unsigned
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        val_one_abs = abs(val_one)
        val_two_abs = abs(val_two)
        val_two_abs = val_two_abs % 32  # only for i32
        result = opfunc.rshift(val_one_abs, val_two_abs)
        check_overflow(result, vm)
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Rotl(Instruction):
    """
        <instr>
             .rotl
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.rotl
        </syntax>
        <descr>
            Integer Rotate Left
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        val_one = abs(val_one)
        val_two = abs(val_two)
        val_one_bin = bin(val_one)[2:]
        diff = 0
        if len(val_one_bin) < 32:
            diff = 32 - len(val_one_bin)
        for i in range(0, diff):
            val_one_bin = "0" + val_one_bin
        val_two = val_two % 32  # only for i32
        result = val_one_bin[val_two:] + val_one_bin[:val_two]
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = int(result, 2)
        vm.inc_sp(line_num)


class Rotr(Instruction):
    """
        <instr>
             .rotr
        </instr>
        <syntax>
            get_local $lhs
            get_local $rhs
            i32.rotr
        </syntax>
        <descr>
            Integer Rotate Right
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        val_one, val_two = get_stack_operands(vm, line_num)
        val_one = abs(val_one)
        val_two = abs(val_two)
        val_one_bin = bin(val_one)[2:]
        diff = 0
        if len(val_one_bin) < 32:
            diff = 32 - len(val_one_bin)
        for i in range(0, diff):
            val_one_bin = "0" + val_one_bin
        val_two = val_two % 32  # only for i32
        first_part = val_one_bin[-val_two:]
        second_part = val_one_bin[: len(val_one_bin) - val_two]
        result = first_part + second_part
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = int(result, 2)
        vm.inc_sp(line_num)


class Clz(Instruction):
    """
        <instr>
             .clz
        </instr>
        <syntax>
            get_local $lhs
            i32.clz
        </syntax>
        <descr>
            Integer Count Leading Zeros
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.dec_sp(line_num)
        val_one = get_stack_operand(vm)
        val_one_bin = bin(val_one)[2:]
        result = 0
        for i in val_one_bin:
            if i == "0":
                result += 1
            else:
                break
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Ctz(Instruction):
    """
        <instr>
             .clz
        </instr>
        <syntax>
            get_local $lhs
            i32.clz
        </syntax>
        <descr>
            Integer Count Trailing Zeros
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.dec_sp(line_num)
        val_one = get_stack_operand(vm)
        val_one_bin = bin(val_one)[2:]
        result = 0
        for i in val_one_bin[::-1]:
            if i == "0":
                result += 1
            else:
                break
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Popcnt(Instruction):
    """
        <instr>
             .popcnt
        </instr>
        <syntax>
            get_local $lhs
            i32.popcnt
        </syntax>
        <descr>
            Integer Population Count
            returns the number of 1-bits in its operand
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.dec_sp(line_num)
        val_one = get_stack_operand(vm)
        val_one_bin = bin(val_one)[2:]
        result = 0
        for i in val_one_bin:
            if i == "1":
                result += 1
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)


class Eqz(Instruction):
    """
        <instr>
             .eqz
        </instr>
        <syntax>
            get_local $lhs
            i32.eqz
        </syntax>
        <descr>
            Integer Equal To Zero
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        vm.dec_sp(line_num)
        val_one = get_stack_operand(vm)
        result = False
        if val_one == 0:
            result = True
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = result
        vm.inc_sp(line_num)



"""
arithmetic.py: arithmetic and logic instructions.
"""

import operator as opfunc
from .argument_check import check_reg_only, check_immediate_three
from assembler.errors import check_num_args, DivisionZero, InvalidArgument

from assembler.tokens import Instruction, MAX_INT, Register


def three_op_arith_reg(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_reg_only(instr, ops, line_num)
    ops[0].set_val(
        check_overflow(
            operator(ops[1].get_val(line_num),
                     ops[2].get_val(line_num)),
            vm),
        line_num)
    vm.changes.add(ops[0].get_nm())


def three_op_arith_immediate(ops, vm, instr, operator, line_num):
    """
        operator: this is the functional version of Python's
            +, -, *, etc.
    """
    check_num_args(instr, ops, 3, line_num)
    check_immediate_three(instr, ops, line_num)
    ops[0].set_val(
        check_overflow(
            operator(ops[1].get_val(line_num),
                     ops[2].get_val(line_num)),
            vm),
        line_num)
    vm.changes.add(ops[0].get_nm())


def check_overflow(val, vm):
    if(val > MAX_INT):
        val = val - MAX_INT+1
    return val


class Add(Instruction):
    """
        <instr>
             ADD
        </instr>
        <syntax>
            ADD reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.add, line_num)


class Addi(Instruction):
    """
        <instr>
             ADDI
        </instr>
        <syntax>
            ADDI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.add, line_num)


class Sub(Instruction):
    """
        <instr>
             SUB
        </instr>
        <syntax>
            SUB reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.sub, line_num)


class Mult(Instruction):
    """
        <instr>
             MULT
        </instr>
        <syntax>
            MULT reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 2, line_num)
        check_reg_only(self.name, ops, line_num)
        result = ops[0].get_val(line_num) * ops[1].get_val(line_num)
        if result > 2 ** 32 - 1:
            vm.registers['LO'] = opfunc.lshift(result, 32)
            vm.registers['HI'] = opfunc.rshift(result, 32)
        else:
            vm.registers['LO'] = result
            vm.registers['HI'] = 0
        vm.changes.add('LO')
        vm.changes.add('HI')
        return ''


class Andf(Instruction):
    """
        <instr>
             AND
        </instr>
        <syntax>
            AND reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.and_, line_num)
        return ''


class Andi(Instruction):
    """
        <instr>
             AND
        </instr>
        <syntax>
            ANDI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.and_, line_num)
        return ''


class Orf(Instruction):
    """
        <instr>
             OR
        </instr>
        <syntax>
            OR reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)
        return ''


class Ori(Instruction):
    """
        <instr>
             OR
        </instr>
        <syntax>
            ORI reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.or_, line_num)
        return ''


class Nor(Instruction):
    """
        <instr>
             NOR
        </instr>
        <syntax>
            NOR reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.or_, line_num)
        ops[0].set_val(opfunc.inv(ops[0].get_val(line_num)), line_num)
        return ''


class Xor(Instruction):
    """
        <instr>
             XOR
        </instr>
        <syntax>
            XOR reg, reg, reg
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_reg(ops, vm, self.name, opfunc.xor, line_num)
        return ''


class Sll(Instruction):
    """
        <instr>
             sll
        </instr>
        <syntax>
            SLL reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.lshift, line_num)
        return ''


class Srl(Instruction):
    """
        <instr>
             srl
        </instr>
        <syntax>
            SRL reg, reg, con
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        three_op_arith_immediate(ops, vm, self.name, opfunc.rshift, line_num)


class Div(Instruction):
    """
        <instr>
             DIV
        </instr>
        <syntax>
            DIV reg, reg
        </syntax>
        <descr>
            The div instruction divides the contents of
            the two registers. The quotient result
            of the division is stored into LO, while the
            remainder is placed in HI.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 2, line_num)
        check_reg_only(self.name, ops, line_num)
        if ops[1].get_val(line_num) == 0:
            raise DivisionZero(line_num)

        quotient = ops[0].get_val(line_num) // ops[1].get_val(line_num)
        remainder = ops[0].get_val(line_num) % ops[1].get_val(line_num)
        vm.registers['LO'] = quotient
        vm.registers['HI'] = remainder
        vm.changes.add('LO')
        vm.changes.add('HI')
        return ''


class Mfhi(Instruction):
    """
        <instr>
             mfhi
        </instr>
        <syntax>
            MFHI reg
        </syntax>
        <descr>
            Moves the value from the HI register into the
            destination register given.
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidArgument(ops[0].get_nm(), line_num)
        ops[0].set_val(vm.registers['HI'], line_num)
        vm.changes.add(ops[0].get_nm())
        return ''


class Mflo(Instruction):
    """
        <instr>
             mflo
        </instr>
        <syntax>
            MFLO reg
        </syntax>
        <descr>
            Moves the value from the LO register into the
            destination register given.
        </descr>
    """

    def fhook(self, ops, vm, line_num):
        check_num_args(self.name, ops, 1, line_num)
        if not isinstance(ops[0], Register):
            raise InvalidArgument(ops[0].get_nm(), line_num)
        ops[0].set_val(vm.registers['LO'], line_num)
        vm.changes.add(ops[0].get_nm())
        return ''



"""
control_flow.py: control flow instructions,
    plus Exceptions to signal break in flow.

"""

from assembler.errors import check_num_args
from assembler.tokens import Instruction
from assembler.flowbreak import Jump
from assembler.ops_check import get_one_op, get_two_ops


class Cmpf(Instruction):
    """
        <instr>
             cmp
        </instr>
        <syntax>
            CMP reg, reg
            CMP reg, mem
            CMP reg, con
        </syntax>
        <descr>
            Compares op1 and op2, and sets (right now) the SF and ZF flags.
            It is not clear at this moment how to
            treat the OF and CF flags in Python,
            since Python integer arithmetic never carries or overflows!
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        (op1, op2) = get_two_ops(self.get_nm(), ops, line_num)
        res = op1.get_val(line_num) - op2.get_val(line_num)
        # set the proper flags
        # zero flag:
        if res == 0:
            vm.flags['ZF'] = 1
        else:
            vm.flags['ZF'] = 0
        # sign flag:
        if res < 0:
            vm.flags['SF'] = 1
        else:
            vm.flags['SF'] = 0
        vm.changes.add('FLAGZF')
        vm.changes.add('FLAGSF')


class Jmp(Instruction):
    """
        <instr>
            jmp
        </instr>
        <syntax>
            JMP lbl
        </syntax>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        raise Jump(target.name)


class Je(Instruction):
    """
        <instr>
             je
        </instr>
        <syntax>
            JE lbl
        </syntax>
        <descr>
            Jumps if ZF is one. <br>
            Equivalent name: JZ
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if int(vm.flags['ZF']) == 1:
            raise Jump(target.name)


class Jne(Instruction):
    """
        <instr>
             jne
        </instr>
        <syntax>
            JNE lbl
        </syntax>
        <descr>
            Jumps if ZF is zero. <br>
            Equivalent name: JNZ
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if int(vm.flags['ZF']) == 0:
            raise Jump(target.name)


class Jg(Instruction):
    """
        <instr>
             jg
        </instr>
        <syntax>
            JG lbl
        </syntax>
        <descr>
            Jumps if SF == 0 and ZF == 0. <br>
            Equivalent name: JLNE
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if (int(vm.flags['SF']) == 0 and
                int(vm.flags['ZF']) == 0):
            raise Jump(target.name)


class Jge(Instruction):
    """
        <instr>
             jge
        </instr>
        <syntax>
            JGE lbl
        </syntax>
        <descr>
            Jumps if SF == 0.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if int(vm.flags['SF']) == 0:
            raise Jump(target.name)
        return ''


class Jl(Instruction):
    """
        <instr>
             jl
        </instr>
        <syntax>
            JL lbl
        </syntax>
        <descr>
            Jumps if SF == 1. <br>
            Equivalent name: JGNE
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if int(vm.flags['SF']) == 1:
            raise Jump(target.name)
        return ''


class Jle(Instruction):
    """
        <instr>
             jle
        </instr>
        <syntax>
            JLE lbl
        </syntax>
        <descr>
            Jumps if SF == 1 or ZF == 1.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        target = get_one_op(self.get_nm(), ops, line_num)
        if (int(vm.flags['SF']) == 1 or
                int(vm.flags['ZF']) == 1):
            raise Jump(target.name)
        return ''


class Call(Instruction):
    """
        <instr>
             call
        </instr>
        <syntax>
            CALL lbl
        </syntax>
        <descr>
            Pushes value of EIP to stack and jumps to the internal subroutine.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("CALL", ops, 1, line_num)
        vm.dec_sp(line_num)
        vm.stack[hex(vm.get_sp() + 1).split('x')[-1].upper()] = vm.get_ip()
        target = get_one_op(self.get_nm(), ops, line_num)
        vm.c_stack.append(vm.get_ip())
        raise Jump(target.name)


class Ret(Instruction):
    """
        <instr>
             ret
        </instr>
        <syntax>
            RET
        </syntax>
        <descr>
            Pops value from stack to EIP and returns control to the
            the line after the subroutine call.
        </descr>
    """
    def fhook(self, ops, vm, line_num):
        check_num_args("RET", ops, 0, line_num)
        vm.inc_sp(line_num)
        vm.set_ip(int(vm.stack[hex(vm.get_sp()).split('x')[-1].upper()]))
        vm.stack[hex(vm.get_sp()).split('x')[-1].upper()] = vm.empty_cell()
        while not isinstance(vm.c_stack[-1], int):
            vm.c_stack.pop()
        if isinstance(vm.c_stack[-1], int):
            vm.c_stack.pop()



"""
assemble.py
Executes assembly code typed in.
"""
from .flowbreak import FlowBreak
from .errors import Error, InvalidInstruction, ExitProg
from .parse import add_debug, parse
from .lex import lex

# from .RISCV.control_flow import  Jr, Jal

MAX_INSTRUCTIONS = 1000  # prevent infinite loops!

JMP_STR = "A jump instruction."

INSTR_INTEL = 0
OPS_INTEL = 1

PC_MIPS = 0
INSTR_MIPS = 1
OPS_MIPS = 2


def dump_flags(vm):
    for flag, val in vm.flags.items():
        add_debug("Flag = " + flag + "; val = " + str(val), vm)


def exec(tok_lines, vm, last_instr):
    """
        Executes a single instruction at location reg[EIP] in tok_lines.
        Returns:
            success: was instruction valid?
            last_instr: any last_instr
            err_msg: if no success, what went wrong?
    """
    try:
        ip = vm.get_ip() - vm.start_ip
        curr_instr = None
        source = None
        last_instr = None

        if vm.past_last_instr(tok_lines):
            raise InvalidInstruction("Past end of code.")
        (curr_instr, source, line_num) = tok_lines[ip // vm.get_ip_div()]
        last_instr = vm.exec_instr(curr_instr, line_num)

        vm.set_next_stack_change()

        return (True, source, "")

    except FlowBreak as brk:
        # we have hit one of the JUMP instructions: jump to that line.
        add_debug("In FlowBreak", vm)
        dump_flags(vm)
        return vm.jump_handler(brk, source, curr_instr)
    except ExitProg:
        raise ExitProg(source)
    except Error as err:
        error_msg = f'Line {err.line_num}: {err.msg}'
        return (False, last_instr, error_msg)


def step_code(tok_lines, vm, error, last_instr, bit_code):
    if vm.get_ip() == 0:
        vm.set_ip(vm.get_start_ip())

    if not vm.past_last_instr(tok_lines):
        (success, last_instr, error) = exec(tok_lines, vm,
                                            last_instr)
    else:
        last_instr = "Reached end of executable code."
        # rewind:
        vm.set_ip(vm.start_ip)

    return (last_instr, error, bit_code)


def run_code(tok_lines, vm, error, last_instr, bit_code):
    count = 0
    add_debug("Setting ip to 0", vm)
    vm.set_ip(vm.get_start_ip())   # instruction pointer reset for 'run'

    while not vm.past_last_instr(tok_lines) and count < MAX_INSTRUCTIONS:
        (success, last_instr, error) = exec(tok_lines, vm,
                                            last_instr)
        if not success:
            break
        count += 1

    if count >= MAX_INSTRUCTIONS:
        error = ("Possible infinite loop detected: "
                 + "instructions run has exceeded " + str(MAX_INSTRUCTIONS))

    return (last_instr, error, bit_code)


def assemble(code, vm, step=False, web=True):
    """
        Assembles and runs code.
        Args:
            code: code to assemble.
            vm:
                Our virtual machine. Contains:
                registers: current register values.
                memory: current memory values.
                flags: current values of flags.
            step: are we stepping through code or running continuously?
        Returns:
            next
            Error, if any.
    """
    last_instr = ''
    error = ''
    bit_code = ''
    if vm.flavor != 'wasm' and vm.next_stack_change != "":
        vm.stack_change = vm.next_stack_change
        vm.next_stack_change = ""
        if len(vm.c_stack) != 0 and not isinstance(vm.c_stack[-1], int):
            vm.c_stack.pop()
        vm.c_stack.append(vm.stack_change)

    if code is None or len(code) == 0:
        return ("", "Must submit code to run.", "")

    tok_lines = None

    # break the code into tokens:
    try:
        tok_lines = lex(code, vm)
        tok_lines = parse(tok_lines, vm, web)

    except Error as err:
        error_msg = f'Line {err.line_num}: {err.msg}'
        return (last_instr, error_msg, bit_code)

    try:
        for curr_instr, source, line_num in tok_lines:
            bit_code += vm.create_bit_instr(curr_instr)

        if step:
            return step_code(tok_lines, vm, error, last_instr, bit_code)
        else:  # step through code
            return run_code(tok_lines, vm, error, last_instr, bit_code)

    except ExitProg as ep:
        last_instr = ep.msg.split(":")[0] + ": Exiting program"
        vm.set_int_ip()
    return (last_instr, error, bit_code)



"""
Our x86 virtual machine representation.
"""

from collections import OrderedDict

from .errors import StackOverflow, StackUnderflow, InvalidArgument
from .MIPS.control_flow import Jal, Jr
from .MIPS.key_words import op_func_codes

MEM_DIGITS = 2

MEM_SIZE = 256
STACK_TOP = (MEM_SIZE * 2) - 1
STACK_BOTTOM = MEM_SIZE
# STACK_TOP = MEM_SIZE - 1
# STACK_BOTTOM = 0
FLOAT_STACK_LIMIT = 8
EMPTY_CELL = 0
INSTR_PTR_INTEL = "EIP"
INSTR_PTR_MIPS = "PC"
INSTR_PTR_RISCV = "PC"
STACK_PTR_INTEL = "ESP"
STACK_PTR_MIPS = "R29"
STACK_PTR_RISCV = "X2"

# Why are these the right values?
MIPS_INTERRUPT_IP = 2147484032
RISC_INTERRUPT_IP = 470351872


class VirtualMachine:
    """
    Holds the memory, registers, flags, etc. that our assembly code
    will use. A singleton class.
    """
    def __init__(self):
        # the x86 registers
        self.nxt_key = 0
        self.ret_str = "GIRONAGIRONAGETSGETS"
        self.debug = ""

        self.memory = OrderedDict()
        self.mem_init()

        self.stack = OrderedDict()
        self.stack_init()

        self.labels = {}
        self.labels_init()

        self.symbols = OrderedDict()
        self.symbols_init()

        self.c_stack = []
        self.cstack_init()

        self.flavor = None
        self.data_init = "on"
        self.start_ip = 0
        self.ip_div = 1
        self.changes = set()
        self.base = None
        self.stack_change = ""
        self.next_stack_change = ""

    def __str__(self):
        return ("Registers: " + str(self.registers) + "\n"
                + "Flags: " + str(self.flags) + "\n"
                + "Memory: " + str(self.memory) + "\n"
                + "Stack: " + str(self.stack) + "\n"
                + "Labels: " + str(self.labels))

    def get_ip_div(self):
        return self.ip_div

    def get_start_ip(self):
        return self.start_ip

    def get_ip(self):
        pass

    def set_ip(self):
        pass

    def past_last_instr(self, token_lines):
        curr_ip = self.get_ip() - self.start_ip
        if curr_ip // self.get_ip_div() >= len(token_lines):
            return True
        return False

    def re_init(self):
        self.nxt_key = 0
        self.ip = 0
        for reg in self.registers:
            self.registers[reg] = 0
        # one gets a unique value:
        for flag in self.flags:
            self.flags[flag] = 0
        self.mem_init()
        self.stack_init()
        self.data_init = "on"
        self.changes_init()
        self.symbols_init()
        self.labels_init()
        self.cstack_init()
        self.stack_change = ""
        self.next_stack_change = ""

    def mem_init(self):
        self.memory.clear()

    def changes_init(self):
        self.changes.clear()

    def symbols_init(self):
        self.symbols.clear()

    def labels_init(self):
        self.labels.clear()

    def cstack_init(self):
        del self.c_stack[:]

    def order_mem(self):
        lst = []
        for key in self.memory:
            lst.append(key)
        for hex_key in range(0, len(lst)):
            lst[hex_key] = int(lst[hex_key], 16)
        lst.sort()
        sorted_mem = OrderedDict()
        for decimal_key in lst:
            hex_sorted_key = hex(decimal_key).split('x')[-1].upper()
            sorted_mem[hex_sorted_key] = self.memory[hex_sorted_key]
        self.memory = sorted_mem

    def empty_cell(self):
        return EMPTY_CELL

    def get_data_init(self):
        return self.data_init

    def set_data_init(self, on_or_off):
        self.data_init = on_or_off

    def jump_to_label(self, label, source, jal=False):
        if label in self.labels:
            ip = self.labels[label]
            self.set_ip(ip + self.start_ip)
            self.next_stack_change = label
            return (True, source, "")
        else:
            try:
                ip = int(label)
                if jal:
                    ip = ip >> 2
                self.set_ip(ip)
                return (True, source, "")
            except Exception:
                return (False, source, f"Invalid label: {label}")

    def jump_handler(self, brk, source, instr_line):
        pass

    def exec_instr(self, instr_line):
        pass

    def create_bit_instr(self, instr_lst):
        return ''

    def set_next_stack_change(self):
        for label in self.labels:
            if self.get_ip() == self.labels[label]:
                self.next_stack_change = label


class IntelMachine(VirtualMachine):
    def __init__(self):
        super().__init__()

        self.float_stack_bottom = -1

        self.registers = OrderedDict(
                    [
                        ('EAX', 0),
                        ('EBX', 0),
                        ('ECX', 0),
                        ('EDX', 0),
                        ('ESI', 0),
                        ('EDI', 0),
                        (STACK_PTR_INTEL, STACK_TOP),
                        ('EBP', 0),
                        (INSTR_PTR_INTEL, 0),
                        ('ST0', 0.0),
                        ('ST1', 0.0),
                        ('ST2', 0.0),
                        ('ST3', 0.0),
                        ('ST4', 0.0),
                        ('ST5', 0.0),
                        ('ST6', 0.0),
                        ('ST7', 0.0),
                    ])

        self.unwritable = [INSTR_PTR_INTEL, STACK_PTR_INTEL]

        # for now we only need four of the flags
        self.flags = OrderedDict(
                    [
                        ('CF', 0),
                        ('OF', 0),
                        ('SF', 0),
                        ('ZF', 0),
                    ])
        self.instr = 0
        self.ops = 1

    def is_FP_stack_empty(self):
        return self.float_stack_bottom == -1

    def is_FP_stack_full(self):
        return self.float_stack_bottom == FLOAT_STACK_LIMIT - 1

    def push_to_Float_Stack(self, val):
        if self.float_stack_bottom == -1:
            self.registers["ST0"] = val
        else:
            for i in range(self.float_stack_bottom + 1, 0, -1):
                self.registers[f'ST{i}'] = self.registers[f'ST{i - 1}']
            self.registers["ST0"] = val
        self.float_stack_bottom += 1
        self.changes.add('ST0')

    def get_register_at_float_stack_top(self):
        return "ST"+str(self.float_stack_top)

    def get_float_stack_register_at_offset(self, k):
        return "ST" + str((self.float_stack_top + k) % FLOAT_STACK_LIMIT)

    def pop_from_Float_Stack(self):
        curr_value_float_stack_top = self.registers['ST0']
        for i in range(0, self.float_stack_bottom):
            self.registers[f'ST{i}'] = self.registers[f'ST{i + 1}']
        self.registers[f'ST{self.float_stack_bottom}'] = 0.0
        self.float_stack_bottom -= 1
        return curr_value_float_stack_top

    def get_next_register(self):
        self.float_stack_top = (self.float_stack_top - 1) % FLOAT_STACK_LIMIT
        return self.float_stack_top

    def reset_FP_Stack(self):
        self.float_stack_bottom = -1
        for i in range(FLOAT_STACK_LIMIT):
            self.registers["ST"+str(i)] = 0.0

    def re_init(self):
        super().re_init()
        self.reset_FP_Stack()
        self.registers[STACK_PTR_INTEL] = STACK_TOP

    def stack_init(self):
        for i in range(STACK_TOP, STACK_BOTTOM - 1, -1):
            self.stack[hex(i).split('x')[-1].upper()] = 0

    def inc_ip(self):
        ip = self.get_ip()
        ip += 1
        self.set_ip(ip)

    def set_ip(self, val):
        self.registers[INSTR_PTR_INTEL] = val

    def set_int_ip(self):
        pass

    def get_ip(self):
        return int(self.registers[INSTR_PTR_INTEL])

    def inc_sp(self, line_num):
        sp = self.get_sp()
        sp += 1
        self.set_sp(sp, line_num)

    def dec_sp(self, line_num):
        sp = self.get_sp()
        sp -= 1
        self.set_sp(sp, line_num)

    def set_sp(self, val, line_num):
        if val < STACK_BOTTOM - 1:
            raise StackOverflow(line_num)
        if val > STACK_TOP:
            raise StackUnderflow(line_num)

        self.registers[STACK_PTR_INTEL] = val

    def get_sp(self):
        return int(self.registers[STACK_PTR_INTEL])

    def jump_handler(self, brk, source, instr_line=None):
        return self.jump_to_label(brk.label, source, True)

    def exec_instr(self, instr_line, line_num):
        self.inc_ip()
        return instr_line[self.instr].f(instr_line[self.ops:], self, line_num)


class MIPSMachine(VirtualMachine):
    """
    Create a VM for running MIPS assembler.
    Why isn't flavor set here?
    """
    def __init__(self):
        super().__init__()
        self.ip_div = 4
        self.unwritable = [INSTR_PTR_MIPS, 'R0', 'F0', 'F29',
                           STACK_PTR_MIPS, 'HI', 'LO']
        self.registers = OrderedDict(
                    [
                        ('R0', 0),
                        ('R12', 0),
                        ('R24', 0),
                        ('R1', 0),
                        ('R13', 0),
                        ('R25', 0),
                        ('R2', 0),
                        ('R14', 0),
                        ('R26', 0),
                        ('R3', 0),
                        ('R15', 0),
                        ('R27', 0),
                        ('R4', 0),
                        ('R16', 0),
                        ('R28', 0),
                        ('R5', 0),
                        ('R17', 0),
                        ('R29', 0),
                        ('R6', 0),
                        ('R18', 0),
                        ('R30', 0),
                        ('R7', 0),
                        ('R19', 0),
                        ('R31', 0),
                        ('R8', 0),
                        ('R20', 0),
                        ('R9', 0),
                        ('R21', 0),
                        ('HI', 0),
                        ('R10', 0),
                        ('R22', 0),
                        ('LO', 0),
                        ('R11', 0),
                        ('R23', 0),
                        (INSTR_PTR_MIPS, 0),
                        ('F0', 0),
                        ('F12', 0),
                        ('F24', 0),
                        ('F1', 0),
                        ('F13', 0),
                        ('F25', 0),
                        ('F2', 0),
                        ('F14', 0),
                        ('F26', 0),
                        ('F3', 0),
                        ('F15', 0),
                        ('F27', 0),
                        ('F4', 0),
                        ('F16', 0),
                        ('F28', 0),
                        ('F5', 0),
                        ('F17', 0),
                        ('F29', 0),
                        ('F6', 0),
                        ('F18', 0),
                        ('F30', 0),
                        ('F7', 0),
                        ('F19', 0),
                        ('F31', 0),
                        ('F8', 0),
                        ('F20', 0),
                        ('F9', 0),
                        ('F21', 0),
                        ('F10', 0),
                        ('F22', 0),
                        ('F11', 0),
                        ('F23', 0)
                    ])

        self.flags = OrderedDict(
                    [
                        ('COND', 0),
                    ])

        self.pc = 0
        self.instr = 1
        self.ops = 2

    def re_init(self):
        super().re_init()
        self.registers[STACK_PTR_MIPS] = STACK_TOP
        self.changes.clear()

    def stack_init(self):
        for i in range(STACK_TOP - 3, STACK_BOTTOM - 1, -4):
            self.stack[hex(i).split('x')[-1].upper()] = 0

    def inc_ip(self):
        ip = self.get_ip()
        ip += 4
        self.set_ip(ip)

    def set_ip(self, val):
        self.registers[INSTR_PTR_MIPS] = val

    def set_int_ip(self):
        self.set_ip(MIPS_INTERRUPT_IP)

    def get_ip(self):
        return int(self.registers[INSTR_PTR_MIPS])

    def inc_sp(self, line_num):
        sp = self.get_sp()
        sp += 1
        self.set_sp(sp, line_num)

    def dec_sp(self, line_num):
        sp = self.get_sp()
        sp -= 1
        self.set_sp(sp, line_num)

    def set_sp(self, val, line_num):
        if val < STACK_BOTTOM - 1:
            raise StackOverflow(line_num)
        if val > STACK_TOP:
            raise StackUnderflow(line_num)

        self.registers[STACK_PTR_MIPS] = val

    def get_sp(self):
        return int(self.registers[STACK_PTR_MIPS])

    def create_bit_negative(self, value, bits):
        """
        Converts an immediate value into a string of bits

        Args:
            value: Immediat value
            bits: Number of bits needed

        Returns:
            Formatted binary value of immediate
            in the number of bits inputted
        """
        imm_code = bin(value).split('b')[1]
        imm_code = '0'*(bits - len(imm_code)) + imm_code
        if value < 0:
            imm_lst = []
            for bit in imm_code:
                imm_lst.append(bit)
            flip_bit = False
            place = bits - 1
            while place >= 0:
                if not flip_bit and imm_lst[place] == "1":
                    flip_bit = True
                elif flip_bit:
                    if imm_lst[place] == "0":
                        imm_lst[place] = "1"
                    else:
                        imm_lst[place] = "0"
                place -= 1
            imm_code = "".join(imm_lst)
        return imm_code

    def create_bit_r_format(self, instr_lst, op_func, func_code):
        """
        Converts an R-format instruction into a string of bits

        Args:
            instr_lst: Line of code

        Returns:
            Formatted version of instruction
        """
        rs = 0
        rt = 0
        shamt = 0
        rd = 0

        # if instruction has shift value
        if func_code == "000000" or func_code == "000010":
            try:
                rd = int(instr_lst[self.ops].get_nm().split('R')[1])
                rt = int(instr_lst[self.ops + 1].get_nm().split('R')[1])
                shamt = instr_lst[self.ops + 2].get_val()
            except Exception:
                pass

        # if function does not use rd
        elif func_code == "011000" or func_code == "011010":
            try:
                rs = int(instr_lst[self.ops].get_nm().split('R')[1])
                rt = int(instr_lst[self.ops + 1].get_nm().split('R')[1])
            except Exception:
                pass

        # if function only uses rd
        elif func_code == "010010" or func_code == "010000":
            try:
                rd = int(instr_lst[self.ops].get_nm().split('R')[1])
            except Exception:
                pass

        # if function only uses rs
        elif func_code == "001000":
            try:
                rs = int(instr_lst[self.ops].get_nm().split('R')[1])
            except Exception:
                pass

        # other arithmetic, logic functions
        else:
            try:
                rs = int(instr_lst[self.ops + 2].get_nm().split('R')[1])
                rt = int(instr_lst[self.ops + 1].get_nm().split('R')[1])
                rd = int(instr_lst[self.ops].get_nm().split('R')[1])
            except Exception:
                pass
        # format the rs, rt, rd, shamt values into 5 bits
        rs = format(rs, '#07b').split('b')[1]
        rt = format(rt, '#07b').split('b')[1]
        rd = format(rd, '#07b').split('b')[1]
        shamt = format(shamt, '#07b').split('b')[1]
        code_lst = [op_func, rs, rt, rd, shamt, func_code, "\n"]
        return " ".join(code_lst)

    def create_bit_i_format(self, instr_lst, op_func):
        """
        Converts an I-format instruction into a string of bits

        Args:
            instr_lst: Line of code

        Returns:
            Formatted version of instruction
        """
        imm = 0
        rs = 0
        rt = 0
        # if not data movement instruction
        if op_func != "100011" and op_func != "101011":
            try:
                rs = int(instr_lst[self.ops + 1].get_nm().split('R')[1])
                rt = int(instr_lst[self.ops].get_nm().split('R')[1])
                imm = int(instr_lst[self.ops + 2].get_val())
            except Exception:
                pass

        # if LW or SW
        else:
            try:
                rs = int(instr_lst[self.ops + 1].get_nm().split('R')[1])
                rt = int(instr_lst[self.ops].get_nm().split('R')[1])
                imm = int(instr_lst[self.ops + 1].displacement)
            except Exception:
                pass

        # format rs, rt into 5 bits
        rs = format(rs, '#07b').split('b')[1]
        rt = format(rt, '#07b').split('b')[1]

        # format imm to 16 bits signed
        imm_code = self.create_bit_negative(imm, 16)
        code_lst = [op_func, rs, rt, imm_code, "\n"]
        return " ".join(code_lst)

    def create_bit_j_format(self, instr_lst, op_func):
        """
        Converts an J-format instruction into a string of bits

        Args:
            instr_lst: Line of code

        Returns:
            Formatted version of instruction
        """
        imm = 0
        try:
            imm = int(instr_lst[self.ops].get_val())
        except Exception:
            pass

        # format imm into 26 bits signed
        imm_code = self.create_bit_negative(imm, 26)
        code_lst = [op_func, imm_code, "\n"]
        return " ".join(code_lst)

    def create_bit_instr(self, instr_lst):
        """
        Converts the instruction into code of bits

        Args:
            instr_lst: Line of code

        Returns:
            Formatted version of the instruction in bits
        """
        op_func = None
        func_code = None
        instr_nm = instr_lst[self.instr].get_nm()
        try:
            op_func, func_code = op_func_codes[instr_nm]
        except Exception:
            op_func = op_func_codes[instr_lst[self.instr].get_nm()]
        if func_code is not None:
            return self.create_bit_r_format(instr_lst, op_func, func_code)
        else:
            if instr_nm != "JAL" and instr_nm != "J":
                return self.create_bit_i_format(instr_lst, op_func)
            else:
                return self.create_bit_j_format(instr_lst, op_func)

    def jump_handler(self, brk, source, instr_line):
        if isinstance(instr_line[self.instr], Jal):
            self.registers["R31"] = self.get_ip()
        if isinstance(instr_line[self.instr], Jr):
            return self.jump_to_label(brk.label, source)
        return self.jump_to_label(brk.label, source, True)

    def exec_instr(self, instr_line, line_num):
        if self.get_ip() != instr_line[self.pc].get_val(line_num):
            raise InvalidArgument(hex(instr_line[self.pc].get_val(line_num)),
                                  line_num)
        self.inc_ip()
        return instr_line[self.instr].f(instr_line[self.ops:], self, line_num)


class RISCVMachine(VirtualMachine):
    # make sure to account for the lack of HI and LO in display

    def __init__(self):
        super().__init__()
        self.ip_div = 4
        self.unwritable = [INSTR_PTR_RISCV, 'X0', STACK_PTR_RISCV]
        self.registers = OrderedDict(
                        [
                            ('X0', 0),
                            ('X11', 0),
                            ('X22', 0),
                            ('X1', 0),
                            ('X12', 0),
                            ('X23', 0),
                            ('X2', 0),
                            ('X13', 0),
                            ('X24', 0),
                            ('X3', 0),
                            ('X14', 0),
                            ('X25', 0),
                            ('X4', 0),
                            ('X15', 0),
                            ('X26', 0),
                            ('X5', 0),
                            ('X16', 0),
                            ('X27', 0),
                            ('X6', 0),
                            ('X17', 0),
                            ('X28', 0),
                            ('X7', 0),
                            ('X18', 0),
                            ('X29', 0),
                            ('X8', 0),
                            ('X19', 0),
                            ('X30', 0),
                            ('X9', 0),
                            ('X20', 0),
                            ('X31', 0),
                            ('X10', 0),
                            ('X21', 0),
                            (INSTR_PTR_RISCV, 0)
                        ])
        self.flags = OrderedDict(
                    [
                        ('COND', 0),
                    ])

        self.pc = 0
        self.instr = 1
        self.ops = 2

    def re_init(self):
        super().re_init()
        self.registers[STACK_PTR_RISCV] = STACK_TOP
        self.changes.clear()

    def stack_init(self):
        for i in range(STACK_TOP - 3, STACK_BOTTOM - 1, -4):
            self.stack[hex(i).split('x')[-1].upper()] = 0

    def inc_ip(self):
        ip = self.get_ip()
        ip += 4
        self.set_ip(ip)

    def set_ip(self, val):
        self.registers[INSTR_PTR_RISCV] = val

    def set_int_ip(self):
        self.set_ip(RISC_INTERRUPT_IP)

    def get_ip(self):
        return int(self.registers[INSTR_PTR_RISCV])

    def inc_sp(self, line_num):
        sp = self.get_sp()
        sp += 1
        self.set_sp(sp, line_num)

    def dec_sp(self, line_num):
        sp = self.get_sp()
        sp -= 1
        self.set_sp(sp, line_num)

    def set_sp(self, val, line_num):
        if val < STACK_BOTTOM - 1:
            raise StackOverflow(line_num)
        if val > STACK_TOP:
            raise StackUnderflow(line_num)

        self.registers[STACK_PTR_RISCV] = val

    def get_sp(self):
        return int(self.registers[STACK_PTR_RISCV])

    def exec_instr(self, instr_line, line_num):
        if self.get_ip() != instr_line[self.pc].get_val(line_num):
            raise InvalidArgument(hex(instr_line[self.pc].get_val(line_num)),
                                  line_num)
        self.inc_ip()
        return instr_line[self.instr].f(instr_line[self.ops:], self, line_num)

    def jump_handler(self, brk, source, instr_line):
        return self.jump_to_label(brk.label, source, True)


class WASMMachine(VirtualMachine):
    def __init__(self):
        super().__init__()
        self.locals = OrderedDict()
        self.locals_init()

        self.globals = OrderedDict()
        self.globals_init()

        self.stack_ptr = STACK_BOTTOM
        self.ip = 0

        self.instr = 0
        self.ops = 1

    def re_init(self):
        self.nxt_key = 0
        self.ip = 0
        self.mem_init()
        self.stack_init()
        self.data_init = "on"
        self.changes_init()
        self.symbols_init()
        self.labels_init()
        self.cstack_init()
        self.locals_init()
        self.globals_init()
        self.stack_change = ""
        self.next_stack_change = ""
        self.stack_ptr = STACK_BOTTOM

    def locals_init(self):
        self.locals.clear()

    def globals_init(self):
        self.globals.clear()

    def stack_init(self):
        for i in range(STACK_TOP - 3, STACK_BOTTOM - 1, -4):
            self.stack[hex(i).split('x')[-1].upper()] = 0

    def inc_sp(self, line_num):
        sp = self.get_sp()
        sp += 4
        self.set_sp(sp, line_num)

    def dec_sp(self, line_num):
        sp = self.get_sp()
        sp -= 4
        self.set_sp(sp, line_num)

    def set_sp(self, val, line_num):
        if val < STACK_BOTTOM - 1:
            raise StackOverflow(line_num)
        if val > STACK_TOP:
            raise StackUnderflow(line_num)

        self.stack_ptr = val

    def get_sp(self):
        return self.stack_ptr

    def inc_ip(self):
        ip = self.get_ip()
        ip += 1
        self.set_ip(ip)

    def set_ip(self, val):
        self.ip = val

    def get_ip(self):
        return self.ip

    def exec_instr(self, instr_line, line_num):
        self.inc_ip()
        return instr_line[self.instr].f(instr_line[self.ops:], self, line_num)

    def set_next_stack_change(self):
        pass


intel_machine = IntelMachine()
mips_machine = MIPSMachine()
riscv_machine = RISCVMachine()
wasm_machine = WASMMachine()



from .arithmetic import Add, Sub, Imul, Idiv, Inc, Dec, Shl
from .arithmetic import Shr, Notf, Andf, Orf, Xor, Neg, BTR, BTS, BSF, BSR
from .arithmetic import BT, BTC
from .fp_arithmetic import FAdd, FSub, FDiv, FMul
from .fp_arithmetic import FAbs, FChs, FaddP, FSubP, FMulP, FDivP, FSqrt
from .control_flow import Cmpf, Je, Jne, Jmp, Call, Ret
from .control_flow import Jg, Jge, Jl, Jle


from .data_mov import Mov, Pop, Push, Lea, Fld, Fst
from .data_mov_att import Movb, Movw, Movl
from .interrupts import Interrupt
from assembler.tokens import DataType, ConstantSign, DupTok

je = Je('JE')
jne = Jne('JNE')
jg = Jg('JG')
jl = Jl('JL')

instructions = {
        # interrupts:
        'INT': Interrupt('INT'),
        # control flow:
        'CMP': Cmpf('CMP'),
        'JMP': Jmp('JMP'),
        je.get_nm(): je,
        jne.get_nm(): jne,
        # the next two instructions are just synonyms for the previous two.
        'JZ': je,
        'JNZ': jne,
        jg.get_nm(): jg,
        jl.get_nm(): jl,
        # JNLE synonymous to JG, JNGE synonymous to JL
        'JNLE': jg,
        'JNGE': jl,
        'JGE': Jge('JGE'),
        'JLE': Jle('JLE'),
        'CALL': Call('CALL'),
        'RET': Ret('RET'),
        # data movement:
        'MOV': Mov('MOV'),
        'PUSH': Push('PUSH'),
        'POP': Pop('POP'),
        'LEA': Lea('LEA'),
        'FLD': Fld('FLD'),
        'FST': Fst('FST'),
        # arithmetic and logic:
        'ADD': Add('ADD'),
        'SUB': Sub('SUB'),
        'IMUL': Imul('IMUL'),
        'IDIV': Idiv('IDIV'),
        'AND': Andf('AND'),
        'OR': Orf('OR'),
        'XOR': Xor('XOR'),
        'SHL': Shl('SHL'),
        'SHR': Shr('SHR'),
        'NOT': Notf('NOT'),
        'INC': Inc('INC'),
        'DEC': Dec('DEC'),
        'NEG': Neg('NEG'),
        'BTR': BTR('BTR'),
        'BTS': BTS('BTS'),
        'BSF': BSF('BSF'),
        'BSR': BSR('BSR'),
        'BT': BT('BT'),
        'BTC': BTC('BTC'),
        # floating point
        'FADD': FAdd('FADD'),
        'FSUB': FSub('FSUB'),
        'FDIV': FDiv('FDIV'),
        'FMUL': FMul('FMUL'),
        'FABS': FAbs('FABS'),
        'FCHS': FChs('FCHS'),
        'FADDP': FaddP('FADDP'),
        'FSUBP': FSubP('FSUBP'),
        'FMULP': FMulP('FMULP'),
        'FDIVP': FDivP('FDIVP'),
        'FSQRT': FSqrt('FSQRT')

        # Other
        }

intel_key_words = {
        # data-types
        'DB': DataType('DB'),
        'DW': DataType('DW'),
        'DD': DataType('DD'),
        # floating point
        "REAL4": DataType('REAL4'),
        "REAL8": DataType('REAL8'),
        "REAL16": DataType('REAL16'),
        "COMPLEX": DataType('COMPLEX'),
        "DUP": DupTok()
}

att_key_words = {
        # other mov instructions
        'MOVB': Movb('MOVB'),
        'MOVW': Movw('MOVW'),
        'MOVL': Movl('MOVL'),

        # data-types
        '.BYTE': DataType('DB'),
        '.SHORT': DataType('DW'),
        '.LONG': DataType('DD'),
        '.FLOAT': DataType('FL'),
        '.DOUBLE': DataType('DBL'),

        # other tokens
        '$': ConstantSign(),
        "DUP": DupTok()
}



class FlowBreak(Exception):
    """
    Base class for all of our flow break exceptions.
    """
    def __init__(self, label):
        super().__init__("Flow break")
        self.label = label
        self.msg = "Unknown control flow exception."


class Jump(FlowBreak):
    def __init__(self, label):
        super().__init__(label)
        self.msg = "Jump to " + label
